/********************************************************************
	created:	2004/02/12
	filename: 	timer\timer.hpp
	author:		Matt Hurd
	
	purpose:	
*********************************************************************/

#ifndef timer_hpp_2004212
#define timer_hpp_2004212

#include <boost/optional.hpp>
#include <sys/timeb.h>
#include <time.h>
#include <vector>
#include <algorithm>
#include <windows.h>
#include <math.h>
#include <boost/lexical_cast.hpp>

#ifdef _DEBUG
	#include <iostream>
#endif

namespace matt {

	std::string time_text ( double time_in_seconds )
	{

		std::string result = "seconds";
		if (abs(time_in_seconds) < 1.0 ) {
			time_in_seconds *= 1000;
			result = "milliseconds";
		}
		if (abs(time_in_seconds) < 1.0 ) {
			time_in_seconds *= 1000;
			result = "microseconds";
		}
		if (abs(time_in_seconds) < 1.0 ) {
			time_in_seconds *= 1000;
			result = "nanoseconds";
		}

		return boost::lexical_cast<std::string>(time_in_seconds) + " " + result;
	}

	class timer 
	{
	public:
		timer()
			: tick_count_(measure())
		{

		}

		timer(double processor_speed_in_GHz)
			: tick_count_(measure())
		{
			#ifdef _DEBUG
				defined_ = true;
			#endif
			processor_speed_in_GHz_ = processor_speed_in_GHz;
		}

		double estimate_frequency( size_t number_of_intervals_of_approximately_a_second = 21 )
		{
			assert( number_of_intervals_of_approximately_a_second > 0);
			std::vector<double>  clock_speed_guess;
			clock_speed_guess.resize(number_of_intervals_of_approximately_a_second);

			__timeb64 timebuffer;
			double first_time, second_time; 
			__int64 first_tick, second_tick;
			double tick_counts_per_second;

			srand( (unsigned)time( NULL ) );

			for (size_t i = 0; i< number_of_intervals_of_approximately_a_second; ++i )
			{
				_ftime64( &timebuffer );
				first_time = double(timebuffer.time) + double(timebuffer.millitm) / 1000.0;
				first_tick = measure();

				Sleep( 800 + long(rand()) * 400 / RAND_MAX );  //  add jitter to the second to avoid tick effects

				second_tick = measure();
				_ftime64( &timebuffer );
				second_time = double(timebuffer.time) + double(timebuffer.millitm) / 1000.0;

				tick_counts_per_second = double(second_tick - first_tick) / (second_time - first_time);
				clock_speed_guess[i] = tick_counts_per_second / 1e9;
			}

			std::sort(clock_speed_guess.begin(), clock_speed_guess.end());

			processor_speed_in_GHz_ = clock_speed_guess[ number_of_intervals_of_approximately_a_second / 2 ];


			#ifdef _DEBUG
				std::cout << "GHz measurement 50% range = ( " 
					<< clock_speed_guess[ number_of_intervals_of_approximately_a_second / 4]
					<< ","
					<< clock_speed_guess[ (number_of_intervals_of_approximately_a_second * 3) / 4]
					<< ") GHz\n";
				std::cout << "GHz median estimate = "
					<< processor_speed_in_GHz_
					<< " GHz\n";

				defined_ = true;
			#endif
			
			return processor_speed_in_GHz_;
		}


		void restart()
		{
			tick_count_ = measure();
		}

		double elapsed()
		{
			__int64 new_count = measure();
			#ifdef _DEBUG
				assert( defined_ );
			#endif

			double elapsed_time = double (new_count - tick_count_) / (processor_speed_in_GHz_ * 1e9);
			tick_count_ = new_count;

			return elapsed_time - (overhead_estimate_? *overhead_estimate_ : 0.0);
		}

		double minimum_overhead_estimator( size_t number_of_trials = 11)
		{
			std::vector<double>  overhead_measure;
			overhead_measure.resize(number_of_trials);

			overhead_estimate_ = boost::optional<double>();

			timer t = *this;
			
			for (size_t i = 0; i< number_of_trials; ++i)
			{
				double time;
				t.restart();
				time = t.elapsed();
				overhead_measure[i] = time;
			}

			std::sort(overhead_measure.begin(), overhead_measure.end());
			#ifdef _DEBUG
				std::cout << "Overhead measurement 50% range = ( " 
					<< time_text(overhead_measure[ number_of_trials/ 4]) 
					<< ","
					<< time_text(overhead_measure[ (number_of_trials * 3) / 4])
					<< ") \n";
				std::cout << "Overhead median estimate = "
					<< time_text(overhead_measure[ number_of_trials / 2])
					<< " \n";
			#endif

			overhead_estimate_.reset( overhead_measure[ number_of_trials / 2] );

			return overhead_measure[ number_of_trials / 2];
		}


	private:
		#ifdef _DEBUG
			static bool	defined_;
		#endif

		static double processor_speed_in_GHz_;
		static boost::optional<double> overhead_estimate_;

		volatile __int64 tick_count_;

	private:

#pragma optimize( "", off )

		__int64 measure()
		{
			volatile ULARGE_INTEGER ts;

			_asm
			{
				xor eax,eax
				push eax
				push ebx
				push ecx
				push edx
				_emit 0x0f					; cpuid - serialise the processor
				_emit 0xa2
				pop edx
				pop ecx
				pop ebx
				pop eax
				_emit 0x0f					; rdtsc
				_emit 0x31
				mov ts.HighPart,edx
				mov ts.LowPart,eax
			}


			return ts.QuadPart;
		}
#pragma optimize( "", on )

	};


}
#endif
